#include "common.h"
#include "Label.h"
#include "DesktopTexture.h"

extern LabelList desktoplabelList;

DesktopTexture::DesktopTexture()
{
	hIcon = 0;

	memoryDC = 0;
	memoryBitmap = 0;
}

DesktopTexture::~DesktopTexture()
{
	if(hIcon)
		DeleteObject(hIcon);

	if(memoryDC)
		DeleteDC(memoryDC);
}

void DesktopTexture::configure(const string &lnkfile, const string &name)
{
	NameValuePair iconTextValues[] = {
		{ "top", 1 },
		{ "bottom", 2 },
		{ "left", 3 },
		{ "right", 4 },
		{ 0, 0 }
	};

	curName = name;

	SHFILEINFO shFileInfo;
    SHGetFileInfo(lnkfile.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON);
	hIcon = shFileInfo.hIcon;
	//hIcon = LoadLSIcon(lnkfile.c_str(), NULL);

	iconSize = GetxRCInt("xlabeldesktop" , "IconSize", 32, 16, 64);
	iconTextPosition = GetxRCNamedValue("xlabeldesktop", "TextPosition", iconTextValues, 2);
	iconTextSpacing = GetxRCInt("xlabeldesktop" , "TextSpacing", 5);
	
	int color;
	char main[32], light[32], dark[32];
	char *buffers[] = {main, light, dark};

	int numTokens = LCTokenize((GetxRCLine("xlabeldesktop", "TextBackground", "ff00ff")).c_str(), buffers, 3, NULL);

	if (numTokens == 3)
	{
		color = strtol(light, NULL, 16);
		iconTextBackgroundLight = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(dark, NULL, 16);
		iconTextBackgroundDark = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	else
	{
		iconTextBackgroundLight = RGB(255,0,255);
		iconTextBackgroundDark = RGB(255,0,255);
	}
	if(numTokens >= 1)
	{
		color = strtol(main, NULL, 16);
		iconTextBackground = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	if (numTokens == 2)
	{
		color = strtol(light, NULL, 16);
		iconTextBackgroundLight = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		iconTextBackgroundDark = iconTextBackgroundLight;
	}

	_saturation = GetxRCInt("xlabeldesktop" , "SaturationIntensity", 255, 0, 255);
	_hueIntensity = GetxRCInt("xlabeldesktop" , "HueIntensity", 0, 0, 255);
	_hueColor = GetxRCColor("xlabeldesktop", "HueColor", RGB(0, 0, 0));
}

void DesktopTexture::apply(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType)
{
	if(hIcon == NULL)
		return;
			
	SetStretchBltMode(hDC, STRETCH_DELETESCANS);

	memoryDC = CreateCompatibleDC(hDC);
	memoryBitmap = CreateCompatibleBitmap(hDC, width, height);
	memoryBitmap = (HBITMAP) SelectObject(memoryDC, memoryBitmap);

	HBRUSH hbrush;
	RECT r;
	r.top = 0;
	r.left = 0;
	r.right = width;
	r.bottom = height;

	hbrush = CreateSolidBrush((RGB(255, 0, 255)));
	FillRect(memoryDC, &r, hbrush);

	int xpos, ypos;
	if (iconTextPosition == 1)
	{
		xpos = (width-iconSize)/2;
		ypos = height-iconSize;
		r.top = 0;
		r.left = 0;
		r.right = width;
		r.bottom = height-iconSize-iconTextSpacing;
	}
	else if (iconTextPosition == 2)
	{
		xpos = (width-iconSize)/2;
		ypos = 0;
		r.top = iconSize+iconTextSpacing;
		r.left = 0;
		r.right = width;
		r.bottom = height;
	}
	else if (iconTextPosition == 3)
	{
		xpos = width-iconSize;
		ypos = (height-iconSize)/2;
		r.top = 0;
		r.left = 0;
		r.right = width-iconSize-iconTextSpacing;
		r.bottom = height;
	}
	else if (iconTextPosition == 4)
	{
		xpos = 0;
		ypos = (height-iconSize)/2;
		r.top = 0;
		r.left = iconSize+iconTextSpacing;
		r.right = width;
		r.bottom = height;
	}

	hbrush = CreateSolidBrush(iconTextBackgroundDark);
	FillRect(memoryDC, &r, hbrush);

	r.right = r.right-1;
	r.bottom = r.bottom-1;
	hbrush = CreateSolidBrush(iconTextBackgroundLight);
	FillRect(memoryDC, &r, hbrush);

	r.left = r.left+1;
	r.top = r.top+1;
	hbrush = CreateSolidBrush(iconTextBackground);
	FillRect(memoryDC, &r, hbrush);
	
	DeleteObject(hbrush);

	int curx = 0;
	int cury = 0;
	for ( LabelListIterator iter = desktoplabelList.begin(); iter != desktoplabelList.end(); iter++)
	{
		if((*iter)->getName() == curName)
		{
			curx = (*iter)->getX();
			cury = (*iter)->getY();
			break;
		}
	}

	PaintDesktopEx(memoryDC, xpos, ypos, iconSize, iconSize, curx+xpos, cury+ypos, false);

	if (hIcon)
	{
		if (_saturation != 255 || _hueIntensity != 0 || imageType == 1)
		{
			ICONINFO ii;
			GetIconInfo(hIcon, &ii);

			HDC hdcBuffer = CreateCompatibleDC(memoryDC);
			SelectObject(hdcBuffer, ii.hbmColor);

			HDC hdcMask = CreateCompatibleDC(hdcBuffer);
			HBITMAP hbmMask = CreateCompatibleBitmap(hdcBuffer, iconSize, iconSize);
			SelectObject(hdcMask, hbmMask);

			// Mask of the icon
			DrawIconEx(hdcMask, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_MASK);

			int iy = 0;
			int ix = 0;
			COLORREF cl = 0;
			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;
			for(iy = 0; iy < iconSize; iy++)
			{
				for(ix = 0; ix < iconSize; ix++)
				{
					// Loop through all transparent pixels...
					// If the mask-pixel is white, then break
					if(!GetPixel(hdcMask, ix, iy))
					{
						cl = GetPixel(hdcBuffer, ix, iy);
						r = GetRValue(cl);
						g = GetGValue(cl);
						b = GetBValue(cl);

						// Saturation effect
						if(_saturation != 255)
						{
							BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

							r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
							g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
							b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
						}

						// Hue effect
						if(_hueIntensity != 0 || imageType == 1)
						{
							int useIntensity = _hueIntensity;
							COLORREF useColor = _hueColor;
							if (imageType == 1)
							{
								useIntensity = 125;
								useColor = RGB(0,0,50);
							}
							// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
							// test normal
							r = (BYTE)((r * (255 - useIntensity) + GetRValue(useColor) * useIntensity + 255) >> 8);
							g = (BYTE)((g * (255 - useIntensity) + GetGValue(useColor) * useIntensity + 255) >> 8);
							b = (BYTE)((b * (255 - useIntensity) + GetBValue(useColor) * useIntensity + 255) >> 8);
						}

						SetPixel(hdcBuffer, ix, iy, RGB(r, g, b));
					}
				}
			}

			StretchBlt( memoryDC, xpos, ypos, iconSize, iconSize, hdcMask, 0, 0, iconSize, iconSize, SRCAND );
			// Combine the foreground with the background
			StretchBlt( memoryDC, xpos, ypos, iconSize, iconSize, hdcBuffer , 0, 0, iconSize, iconSize, SRCPAINT );
			
			DeleteObject(hbmMask);
			DeleteDC(hdcMask);
			DeleteDC(hdcBuffer);
		}
		else
			DrawIconEx(memoryDC, xpos, ypos, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);
	}

	BitBlt( hDC, x, y, width, height, memoryDC, 0, 0, SRCCOPY );
}


