#include "common.h"
#include "LabelSettings.h"
#include "Font.h"
#include "Texture.h"

NameValuePair alignValues[] = {
	{ "left", DT_LEFT },
	{ "center", DT_CENTER },
	{ "right", DT_RIGHT },
	{ 0, 0 }
};

NameValuePair vertAlignValues[] = {
	{ "top", ALIGN_TOP },
	{ "center", ALIGN_CENTER },
	{ "bottom", ALIGN_BOTTOM },
	{ 0, 0 }
};

NameValuePair keyValues[] = {
	{ "ctrl", 1 },
	{ "shift", 2 },
	{ "ctrl+shift", 3 },
	{ "shift+ctrl", 3 },
	{ 0, 0 }
};

NameValuePair buttonValues[] = {
	{ "left", 1 },
	{ "right", 2 },
	{ "middle", 3 },
	{ 0, 0 }
};

NameValuePair autoWidthValues[] = {
	{ "left", 1 },
	{ "right", 2 },
	{ "center", 3 },
	{ 0, 0 }
};

NameValuePair autoHeightValues[] = {
	{ "top", 1 },
	{ "bottom", 2 },
	{ "center", 3 },
	{ 0, 0 }
};

NameValuePair transparencyValues[] = {
	{ "true", 1 },
	{ "fake", 2 },
	{ "false", 3 },
	{ "off", 3 },
	{ "none", 3 },
	{ ".none", 3 },
	{ 0, 0 }
};

NameValuePair effectValues[] = {
	{ "Random", -1},
	{ "Spin", 0},
	{ "Vortex", 1},
	{ "ScatterGather", 2},
	{ "Spike", 3},
	{ "Fireworks", 4},
	{ 0, 0 }
};

NameValuePair overeffectValues[] = {
	{ "Random", -1},
	{ "None", 0},
	{ "VerticalLines", 1},
	{ "HorizontalLines", 2},
	{ "Switch(lr)", 3},
	{ "Switch(rl)", 4},
	{ "Wave(tb)", 5},
	{ "Wave(bt)", 6},
	{ "Wave(lr)", 7},
	{ "Wave(rl)", 8},
	{ "Pixelize", 9},
	{ "Border", 10},
	{ "Stretch(lr)", 11},
	{ "Stretch(rl)", 12},
	{ "Stretch(tb)", 13},
	{ "Stretch(bt)", 14},
	{ "Slide(lr)", 15},
	{ "Slide(rl)", 16},
	{ "Slide(tb)", 17},
	{ "Slide(bt)", 18},
	{ "Slide(ltrb)", 19},
	{ "Slide(lbrt)", 20},
	{ 0, 0 }
};

NameValuePair scrollValues[] = {
	{ "horizontal", 1},
	{ "vertical", 2},
	{ "horizontal-left", 1},
	{ "vertical-up", 2},
	{ "horizontal-right", 3},
	{ "vertical-down", 4},
	{ 0, 0 }
};

LabelSettings::LabelSettings()
{
	const char *name = "AllLabels";
	isDefault = true;

	labelGroup = GetxRCString(name, "AddToGroup", "");

	alwaysOnTop = GetxRCBool(name, "AlwaysOnTop");
	startHidden = GetxRCBool(name, "StartHidden");
	bUseFahrenheit = GetxRCBool(name, "UseFahrenheit");

	skin = GetRCTexture(name, "");

	font = GetRCFont(name);
	
	leftBorder = GetxRCInt(name, "LeftBorder", 0);
	topBorder = GetxRCInt(name, "TopBorder", 0);
	rightBorder = GetxRCInt(name, "RightBorder", 0);
	bottomBorder = GetxRCInt(name, "BottomBorder", 0);

	align = GetxRCNamedValue(name, "Align", alignValues, DT_CENTER);
	vertAlign = GetxRCNamedValue(name, "VertAlign", vertAlignValues, ALIGN_CENTER);
	updateInterval = GetxRCInt(name, "UpdateInterval", 1000);
	alwaysUpdateContent = GetxRCBool(name, "UpdateAlways");

	int screenX = GetSystemMetrics(SM_CXSCREEN);
	int screenY = GetSystemMetrics(SM_CYSCREEN);

	width = GetxRCDimen(name, "Width", 64, screenX);
	height = GetxRCDimen(name, "Height", 64, screenY);

	x = GetxRCCoord(name, "X", 0, screenX);
	y = GetxRCCoord(name, "Y", 0, screenY);

	text = GetxRCString(name, "Text", "");
	tooltip = GetxRCString(name, "Tooltip", "");
	tooltipmaxwidth = GetxRCInt(name, "TooltipMaxWidth", 300, 0);

	smooth = GetxRCBool(name, "FontSmoothing", true);

	focusOnEnter = GetxRCBool(name, "FocusOnEnter");

	lineBreak = GetxRCBool(name, "AutoLineBreak");

	leftClickCommand = GetxRCLine(name, "OnLeftClick", "");
	leftDoubleClickCommand = GetxRCLine(name, "OnLeftDoubleClick", "");
	middleClickCommand = GetxRCLine(name, "OnMiddleClick", "");
	middleDoubleClickCommand = GetxRCLine(name, "OnMiddleDoubleClick", "");
	rightClickCommand = GetxRCLine(name, "OnRightClick", "");
	rightDoubleClickCommand = GetxRCLine(name, "OnRightDoubleClick", "");
	wheelDownCommand = GetxRCLine(name, "OnWheelDown", "");
	wheelUpCommand = GetxRCLine(name, "OnWheelUp", "");
	enterCommand = GetxRCLine(name, "OnEnter", "");
	leaveCommand = GetxRCLine(name, "OnLeave", "");
	dropCommand = GetxRCLine(name, "OnDrop", "");
	textChangeCommand = GetxRCLine(name, "OnTextChange", "");
	resizeCommand = GetxRCLine(name, "OnResize", "");

	scrollPadLength = GetxRCInt(name, "scrollPad", 10); 
	scrollInterval = GetxRCInt(name, "scrollInterval", 100);
	scrollSpeed = GetxRCInt(name, "scrollSpeed", 1);
	scroll = GetxRCNamedValue(name, "scroll", scrollValues, 0);

	autoWidthMode = GetxRCNamedValue(name, "AutoWidthMode", autoWidthValues, 0);
	autoHeightMode = GetxRCNamedValue(name, "AutoHeightMode", autoHeightValues, 0);
	autoMaxWidth = GetxRCDimen(name, "AutoMaxWidth", screenX, screenX);
	autoMaxHeight = GetxRCDimen(name, "AutoMaxHeight", screenY, screenY);

	transparencyMode = GetxRCNamedValue(name, "TransparencyMode", transparencyValues, 1);
	alphaTransparency = GetxRCInt(name, "AlphaTransparency", 255);
	alphaFade = GetxRCBool(name, "AlphaFade");
	ghosted = GetxRCBool(name, "Ghosted");

	moveable = GetxRCBool(name, "Moveable");
	moveKey = GetxRCNamedValue(name, "MoveModifierKey", keyValues, 1);
	moveButton = GetxRCNamedValue(name, "MoveButton", buttonValues, 1);

	string helper;
	helper = "*";
	helper.append(name);
	helper.append("OnLeftClickRegion");
	labelLeftClickRegions = GetxRCLineList(helper.c_str());
	helper = "*";
	helper.append(name);
	helper.append("OnRightClickRegion");
	labelRightClickRegions = GetxRCLineList(helper.c_str());
	helper = "*";
	helper.append(name);
	helper.append("OnMiddleClickRegion");
	labelMiddleClickRegions = GetxRCLineList(helper.c_str());

	helper = "*";
	helper.append(name);
	helper.append("ModuleHook");
	moduleHook = GetxRCLineList(helper.c_str());

	configSavePath = GetxRCLine(name, "ConfigFile", "");
	cursorPath = GetxRCString(name, "Cursor", "");

	//Animations
	imageLoop = GetxRCInt(name, "FrameLoops", -1);
	imageFrameDelay = GetxRCInt(name, "FrameDelay", 100);
	imageFrameCount = GetxRCInt(name, "FrameCount", 0);
	hoverimageLoop = GetxRCInt(name, "HoverFrameLoops", -1);
	hoverimageFrameDelay = GetxRCInt(name, "HoverFrameDelay", 100);
	hoverimageFrameCount = GetxRCInt(name, "HoverFrameCount", 0);

	overblendStyle = GetxRCNamedValue(name, "OverBlendEffect", overeffectValues, 0);

	//Plugins
	htmltext = GetxRCBool(name, "PluginHtmlText");
	anim = GetxRCBool(name, "PluginAnim");
	animEffect = GetxRCNamedValue(name, "AnimEffect", effectValues, 1);
	animSteps = GetxRCInt(name, "AnimSteps", 50);
	animDelay = GetxRCInt(name, "AnimDelay", 5);
	animAfterSteps = GetxRCInt(name, "AnimAfterSteps", 6);
}

LabelSettings::LabelSettings(const char *name, HWND box)
{
	isDefault = false;

	labelGroup = GetxRCString(name, "AddToGroup", defaultSettings->labelGroup);

	//Create and check for labelgroupdefault settings
	LabelSettings* defaultgroupSettings = NULL;
	if (!labelGroup.empty())
		defaultgroupSettings = new LabelSettings( labelGroup.c_str(), box );
	else
		defaultgroupSettings = defaultSettings;

	//-------------------------------------------------------------------------------------
	alwaysOnTop = GetxRCBool(name, "AlwaysOnTop", defaultgroupSettings->alwaysOnTop);
	startHidden = GetxRCBool(name, "StartHidden", defaultgroupSettings->startHidden);
	bUseFahrenheit = GetxRCBool(name, "UseFahrenheit", defaultgroupSettings->bUseFahrenheit);

	skin = GetRCTexture(name, "", defaultgroupSettings->skin);

	font = GetRCFont(name, defaultgroupSettings->font);
	
	leftBorder = GetxRCInt(name, "LeftBorder", defaultgroupSettings->leftBorder);
	topBorder = GetxRCInt(name, "TopBorder", defaultgroupSettings->topBorder);
	rightBorder = GetxRCInt(name, "RightBorder", defaultgroupSettings->rightBorder);
	bottomBorder = GetxRCInt(name, "BottomBorder", defaultgroupSettings->bottomBorder);

	align = GetxRCNamedValue(name, "Align", alignValues, defaultgroupSettings->align);
	vertAlign = GetxRCNamedValue(name, "VertAlign", vertAlignValues, defaultgroupSettings->vertAlign);
	updateInterval = GetxRCInt(name, "UpdateInterval", defaultgroupSettings->updateInterval);
	alwaysUpdateContent = GetxRCBool(name, "UpdateAlways", defaultgroupSettings->alwaysUpdateContent);

	int maxX;
	int maxY;
	if (box != NULL)
	{
		RECT rc;
		GetClientRect(box, &rc);
		maxX = rc.right-rc.left;
		maxY = rc.bottom-rc.top;
	}
	else
	{
		maxX = GetSystemMetrics(SM_CXSCREEN);
		maxY = GetSystemMetrics(SM_CYSCREEN);
	}

	width = GetxRCDimen(name, "Width", defaultgroupSettings->width, maxX);
	height = GetxRCDimen(name, "Height", defaultgroupSettings->height, maxY);
	x = GetxRCCoord(name, "X", defaultgroupSettings->x, maxX);
	y = GetxRCCoord(name, "Y", defaultgroupSettings->y, maxY);

	text = GetxRCString(name, "Text", defaultgroupSettings->text);
	tooltip = GetxRCString(name, "Tooltip", defaultgroupSettings->tooltip);
	tooltipmaxwidth = GetxRCInt(name, "ToolTipMaxWidth", defaultgroupSettings->tooltipmaxwidth);

	smooth = GetxRCBool(name, "FontSmoothing", defaultgroupSettings->smooth);

	focusOnEnter = GetxRCBool(name, "FocusOnEnter", defaultgroupSettings->focusOnEnter);

	lineBreak = GetxRCBool(name, "AutoLineBreak", defaultgroupSettings->lineBreak);

	leftClickCommand = GetxRCLine(name, "OnLeftClick", defaultgroupSettings->leftClickCommand);
	leftDoubleClickCommand = GetxRCLine(name, "OnLeftDoubleClick", defaultgroupSettings->leftDoubleClickCommand);
	middleClickCommand = GetxRCLine(name, "OnMiddleClick", defaultgroupSettings->middleClickCommand);
	middleDoubleClickCommand = GetxRCLine(name, "OnMiddleDoubleClick", defaultgroupSettings->middleDoubleClickCommand);
	rightClickCommand = GetxRCLine(name, "OnRightClick", defaultgroupSettings->rightClickCommand);
	rightDoubleClickCommand = GetxRCLine(name, "OnRightDoubleClick", defaultgroupSettings->rightDoubleClickCommand);
	wheelDownCommand = GetxRCLine(name, "OnWheelDown", defaultgroupSettings->wheelDownCommand);
	wheelUpCommand = GetxRCLine(name, "OnWheelUp", defaultgroupSettings->wheelUpCommand);
	enterCommand = GetxRCLine(name, "OnEnter", defaultgroupSettings->enterCommand);
	leaveCommand = GetxRCLine(name, "OnLeave", defaultgroupSettings->leaveCommand);
	dropCommand = GetxRCLine(name, "OnDrop", defaultgroupSettings->dropCommand);
	textChangeCommand = GetxRCLine(name, "OnTextChange", defaultgroupSettings->textChangeCommand);
	resizeCommand = GetxRCLine(name, "OnResize", defaultgroupSettings->resizeCommand);

	scrollPadLength = GetxRCInt(name, "scrollPad", defaultgroupSettings->scrollPadLength);
	scrollInterval = GetxRCInt(name, "scrollInterval", defaultgroupSettings->scrollInterval);
	scrollSpeed = GetxRCInt(name, "scrollSpeed", defaultgroupSettings->scrollSpeed);
	scroll = GetxRCNamedValue(name, "scroll", scrollValues, defaultgroupSettings->scroll);
	//Compatibility to old Boolean Value Scroll
	if (GetxRCBool(name, "scroll") && scroll == 0)
		scroll = 1;

	autoWidthMode = GetxRCNamedValue(name, "AutoWidthMode", autoWidthValues, defaultgroupSettings->autoWidthMode);
	autoHeightMode = GetxRCNamedValue(name, "AutoHeightMode", autoHeightValues, defaultgroupSettings->autoHeightMode);
	autoMaxWidth = GetxRCDimen(name, "AutoMaxWidth", defaultgroupSettings->autoMaxWidth, maxX);
	autoMaxHeight = GetxRCDimen(name, "AutoMaxHeight", defaultgroupSettings->autoMaxHeight, maxY);

	transparencyMode = GetxRCNamedValue(name, "TransparencyMode", transparencyValues, defaultgroupSettings->transparencyMode);
	alphaTransparency = GetxRCInt(name, "AlphaTransparency", defaultgroupSettings->alphaTransparency);
	alphaFade = GetxRCBool(name, "AlphaFade", defaultgroupSettings->alphaFade);
	ghosted = GetxRCBool(name, "Ghosted", defaultgroupSettings->ghosted);

	moveable = GetxRCBool(name, "Moveable", defaultgroupSettings->moveable);
	moveKey = GetxRCNamedValue(name, "MoveModifierKey", keyValues, defaultgroupSettings->moveKey);
	moveButton = GetxRCNamedValue(name, "MoveButton", buttonValues, defaultgroupSettings->moveButton);

	string helper;
	helper = "*";
	helper.append(name);
	helper.append("OnLeftClickRegion");
	labelLeftClickRegions = GetxRCLineList(helper.c_str());
	helper = "*";
	helper.append(name);
	helper.append("OnRightClickRegion");
	labelRightClickRegions = GetxRCLineList(helper.c_str());
	helper = "*";
	helper.append(name);
	helper.append("OnMiddleClickRegion");
	labelMiddleClickRegions = GetxRCLineList(helper.c_str());

	helper = "*";
	helper.append(name);
	helper.append("ModuleHook");
	moduleHook = GetxRCLineList(helper.c_str());

	configSavePath = GetxRCLine(name, "ConfigFile", defaultgroupSettings->configSavePath);
	cursorPath = GetxRCString(name, "Cursor", defaultgroupSettings->cursorPath);

	//Animations
	imageLoop = GetxRCInt(name, "FrameLoop", defaultgroupSettings->imageLoop);
	imageFrameDelay = GetxRCInt(name, "FrameDelay", defaultgroupSettings->imageFrameDelay);
	imageFrameCount = GetxRCInt(name, "FrameCount", defaultgroupSettings->imageFrameCount);
	hoverimageLoop = GetxRCInt(name, "HoverFrameLoop", defaultgroupSettings->hoverimageLoop);
	hoverimageFrameDelay = GetxRCInt(name, "HoverFrameDelay", defaultgroupSettings->hoverimageFrameDelay);
	hoverimageFrameCount = GetxRCInt(name, "HoverFrameCount", defaultgroupSettings->hoverimageFrameCount);

	overblendStyle = GetxRCNamedValue(name, "OverblendEffect", overeffectValues, defaultgroupSettings->overblendStyle);

	//Plugins
	htmltext = GetxRCBool(name, "PluginHtmlText", defaultgroupSettings->htmltext);
	anim = GetxRCBool(name, "PluginAnim", defaultgroupSettings->anim);
	animEffect = GetxRCNamedValue(name, "AnimEffect", effectValues, defaultgroupSettings->animEffect);
	animSteps = GetxRCInt(name, "AnimSteps", defaultgroupSettings->animSteps, 1, 255);
	animDelay = GetxRCInt(name, "AnimDelay", defaultgroupSettings->animDelay, 0, 100);
	animAfterSteps = GetxRCInt(name, "AnimAfterSteps", defaultgroupSettings->animAfterSteps, 0, 32);
	//-------------------------------------------------------------------------------------

	if (defaultgroupSettings != defaultSettings)
		delete defaultgroupSettings;
}

LabelSettings::~LabelSettings()
{
	if(isDefault)
	{
		delete skin;
		delete font;
	}
}

LabelSettings* defaultSettings = NULL;
