#include "common.h"
#include "Label.h"
#include "LabelSettings.h"
#include "Texture.h"
#include "Font.h"
#include "SystemInfo.h"

#define _MERGE_RDATA_
#include "AggressiveOptimize.h"

// Bang commands
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangDebug(HWND hwndCaller, LPCSTR pszArgs);
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangGrow(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs);
void BangMove(HWND hwndCaller, LPCSTR pszArgs);
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs);
void BangNext(HWND hwndCaller, LPCSTR pszArgs);
void BangPinToDesktop(HWND hwndCaller, LPCSTR pszArgs);
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs);
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs);
void BangReposition(HWND hwndCaller, LPCSTR pszArgs);
void BangResize(HWND hwndCaller, LPCSTR pszArgs);
void BangScroll(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs);
void BangSetFontColor(HWND hwndCaller, LPCSTR pszArgs);
void BangSetFontShadowColor(HWND hwndCaller, LPCSTR pszArgs);
void BangSetText(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs);
void BangShrink(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);
void BangToggleAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangToggleGhosted(HWND hwndCaller, LPCSTR pszArgs);
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs);

void BangmzscriptVarCopy(HWND hwndCaller, LPCSTR pszArgs);
void BangPosmzscriptVarCopy(HWND hwndCaller, LPCSTR pszArgs);
void BangSizemzscriptVarCopy(HWND hwndCaller, LPCSTR pszArgs);

void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs);
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs);

//Desktop
void BangDesktopCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangDesktopSave(HWND hwndCaller, LPCSTR pszArgs);

StringList getActualDesktopItemPaths(bool defaultDesktop, string searchPath);
void DesktopUpdate(int updateMode, string updatedFile, StringList ActualDesktopItemPaths);

//DesktopWatcherPlugin
typedef VOID (*START)(LPCSTR path);
typedef VOID (*STOP)(LPCSTR path = NULL);

static HMODULE m_PluginDesktopWatcher = NULL;

static START pStart = NULL;
static STOP pStop = NULL;

//Textedit-2.52
#include "regexp.h"
#include "regmagic.h"

void getToken(char line[4096], char temp[4096], bool escape_codes = true);

void BangTextEdittextAppend(HWND hwndCaller, LPCSTR pszArgs);
void BangTextEdittextDelete(HWND hwndCaller, LPCSTR pszArgs);
void BangTextEdittextReplace(HWND hwndCaller, LPCSTR pszArgs);

//Shortcut 2
void BangShortcutHide(HWND hwndCaller, LPCSTR pszArgs);
void BangShortcutShow(HWND hwndCaller, LPCSTR pszArgs);
void BangShortcutToggle(HWND hwndCaller, LPCSTR pszArgs);
void BangShortcutOnBottom(HWND hwndCaller, LPCSTR pszArgs);
void BangShortcutOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangShortcutOnTopToggle(HWND hwndCaller, LPCSTR pszArgs);

bool usetextedit = GetxRCBool("xlabel", "UseTextEdit");
bool useshortcut2 = GetxRCBool("xlabel", "UseShortcut2");

bool useautodesktop = GetxRCBool("xlabeldesktop", "AutoCreate");
string desktopSavePath = GetxRCString("xlabeldesktop", "ConfigFile", "");

string Bang = "";

//LPCTSTR szLogName = "xLabel";

//LSLog(LOG_NOTICE, szLogName, "Message");
//LSLog(LOG_DEBUG, szLogName, "Message");
//LSLog(LOG_WARNING, szLogName, "Message");
//LSLog(LOG_ERROR, szLogName, "Message");
//LSLogPrintf(LOG_DEBUG, szLogName, "Hidden icons (%d):", hiddenIcons.size());

HINSTANCE hInstance;
HWND messageHandler;

LabelList labelList;
LabelList desktoplabelList;
LabelList shortcutlabelList;

StringList labelDesktopItems;
StringList labelShortcutItems;

int lsMessages[] = {
	LM_GETREVID,
	LM_REFRESH,
	0
};

#define LM_UPDATEBG (WM_USER + 1)

Label *lookupLabel( const string &name, LabelList searchList = labelList );

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}
		
		case LM_REFRESH:
		{
			StringList labelNames = GetxRCNameList("*label");

			// refresh the "AllLabels" configuration
			delete defaultSettings;
			defaultSettings = new LabelSettings();

			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if(!(*lit)->getBox())
				{
					// destroy all labels that no longer exist and that are not in a box
					for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
					{
						if(stricmp((*sit).c_str(), (*lit)->getName().c_str()) == 0)
							break;
					}
					if (sit == labelNames.end())
					{
						labelList.remove(*lit);
						delete *lit;
						continue;
					}
				}

				// we can reconfigure all other labels, even if they are "boxed"
				(*lit)->reconfigure();
				(*lit)->repaint(true);
			}

			// create the rest
			for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
			{
				Label *label = lookupLabel(*sit);
				
				if (!label) 
				{
					label = new Label(*sit);

					label->load(hInstance);
				
					labelList.insert(labelList.end(), label);
				}
			}
			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			LabelListIterator lit;
			for(lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				Label *label = *lit;

				if(label->getBox() == 0 && label->getTransparencyMode() == 2)
					label->repaint(true);
			}

			for(lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				Label *label = *lit;

				if(label->getBox() == 0)
					label->repaint(true);
			}

			return 0;
		}

		case WM_DISPLAYCHANGE:

		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return 0;
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS;
	wc.lpfnWndProc = Label::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Label *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabel";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabelMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xLabelMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	defaultSettings = new LabelSettings();
	systemInfo = new SystemInfo();

	StringList labelNames = GetxRCNameList("*label");

	StringListIterator sit;

	for(sit = labelNames.begin(); sit != labelNames.end(); sit++)
	{
		Label *label = new Label(*sit);
	
		labelList.insert(labelList.end(), label);
		
		label->load(hInstance);
	}

	//Desktop Items
	labelDesktopItems = GetxRCLineList("*xlabeldesktopitem");

	if (!labelDesktopItems.empty())
	{	
		for(sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
		{
			char name[32];
			char *buffers[] = {name};

			if(LCTokenize((*sit).c_str(), buffers, 1, 0) >= 1)
			{
				Label *label = new Label(name);

				desktoplabelList.insert(desktoplabelList.end(), label);

				label->desktopload(hInstance, (*sit));
			}
		}
	}
	 
	//Shortcut Items
	if (useshortcut2)
	{
		labelShortcutItems = GetxRCLineList("*shortcut");

		if (!labelShortcutItems.empty())
		{
			int lastx = 0;
			int lasty = 0;
			for(sit = labelShortcutItems.begin(); sit != labelShortcutItems.end(); sit++)
			{
				char name[64], ix[16], iy[16];
				char *buffers[] = {name, ix, iy};

				if(LCTokenize((*sit).c_str(), buffers, 3, 0) >= 3)
				{
					Label *label = new Label(name);

					shortcutlabelList.insert(shortcutlabelList.end(), label);

					label->shortcutload(hInstance, (*sit), lastx, lasty);

					lastx = xParseCoord(ix, 0, GetSystemMetrics(SM_CXSCREEN), lastx);
					lasty = xParseCoord(iy, 0, GetSystemMetrics(SM_CYSCREEN), lasty);		
				}
			}
		}
	}

	AddBangCommand("!LabelAlwaysOnTop", BangAlwaysOnTop);
	AddBangCommand("!LabelClipboardCopy", BangClipboardCopy);
	AddBangCommand("!LabelClipboardPaste", BangClipboardPaste);
	AddBangCommand("!LabelCreate", BangCreate);
	AddBangCommand("!LabelDebug", BangDebug);
	AddBangCommand("!LabelDestroy", BangDestroy);
	AddBangCommand("!LabelHide", BangHide);
	AddBangCommand("!LabelLSBoxHook", BangLSBoxHook);
	AddBangCommand("!LabelMove", BangMove);
	AddBangCommand("!LabelNext", BangNext);
	AddBangCommand("!LabelPinToDesktop", BangPinToDesktop);
	AddBangCommand("!LabelPrevious", BangPrevious);
	AddBangCommand("!LabelRefresh", BangRefresh);
	AddBangCommand("!LabelReposition", BangReposition);
	AddBangCommand("!LabelResize", BangResize);
	AddBangCommand("!LabelScroll", BangScroll);
	AddBangCommand("!LabelSetFontColor", BangSetFontColor);
	AddBangCommand("!LabelSetFontShadowColor", BangSetFontShadowColor);
	AddBangCommand("!LabelSetText", BangSetText);
	AddBangCommand("!LabelShow", BangShow);
	AddBangCommand("!LabelShowHide", BangShowHide);
	AddBangCommand("!LabelToggleAlwaysOnTop", BangToggleAlwaysOnTop);
	AddBangCommand("!LabelToggle", BangToggle);
	AddBangCommand("!LabelUpdate", BangUpdate);
	if(IsOS(OS_2KXP))
		AddBangCommand("!LabelSetAlpha", BangSetAlpha);
	AddBangCommand("!LabelToggleGhosted", BangToggleGhosted);

	AddBangCommand("!LabelmzscriptVarCopy", BangmzscriptVarCopy);
	AddBangCommand("!LabelPosmzscriptVarCopy", BangPosmzscriptVarCopy);
	AddBangCommand("!LabelSizemzscriptVarCopy", BangSizemzscriptVarCopy);

	AddBangCommand("!LabelSetAnimation", BangSetAnimation);

	AddBangCommand("!LabelShrink", BangShrink);
	AddBangCommand("!LabelGrow", BangGrow);
	AddBangCommand("!LabelMoveBy", BangMoveBy);

	AddBangCommand("!ParseEvars", BangParseEvars);
	AddBangCommand("!SetEvar", BangSetEvars);

	//Desktop
	AddBangCommand("!LabelDesktopCreate", BangDesktopCreate);
	AddBangCommand("!LabelDesktopSave", BangDesktopSave);

	//TextEdit 2.52
	if (usetextedit)
	{
		//LSLog(LOG_NOTICE, szLogName, "Textedit-2.52 Bangs loaded!");
		AddBangCommand("!textappend", BangTextEdittextAppend);
		AddBangCommand("!textdel", BangTextEdittextDelete);
		AddBangCommand("!textreplace", BangTextEdittextReplace);
	}

	//Shortcut2
	if (useshortcut2)
	{
		//LSLog(LOG_NOTICE, szLogName, "Shortcut2 Bangs loaded!");
		AddBangCommand("!ShortcutGroupHide", BangShortcutHide);
		AddBangCommand("!ShortcutGroupShow", BangShortcutShow);
		AddBangCommand("!ShortcutGroupToggle", BangShortcutToggle);
		AddBangCommand("!ShortcutGroupOnTop", BangShortcutOnTop);
		AddBangCommand("!ShortcutGroupOnBottom", BangShortcutOnBottom);
		AddBangCommand("!ShortcutGroupOnTopToggle", BangShortcutOnTopToggle);
	}

	//ModuleHook and
	//OnResize one time after all are loaded
	for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		if (!(*lit)->moduleHook.empty())
		{
			for(StringListIterator sit = (*lit)->moduleHook.begin(); sit != (*lit)->moduleHook.end(); sit++)
			{
				string hWndbuffer = IntToString( (int) (*lit)->gethWnd() );
				string config = (*sit).c_str();

				if ( !(*lit)->configSavePath.empty() )
					config.append( " \"" + (*lit)->configSavePath + "\" " + hWndbuffer );
				else
					config.append( " \"$litestepdir$\" " + hWndbuffer);

				LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
			}
		}

		if( !(*lit)->resizeCommand.empty() )
			LSExecute(NULL, (*lit)->resizeCommand.c_str(), SW_SHOWNORMAL);
	}

	if (useautodesktop)
	{
		if (labelDesktopItems.empty())
		{
			Bang = "!labeldesktopcreate replace \".desktop\"";
			LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
		}

		string pluginName = "xdesktopwatcher.dll";

		m_PluginDesktopWatcher = LoadLibrary(pluginName.c_str());

		if(m_PluginDesktopWatcher == NULL)
		{
			char tmpSz[4096];
			// Try to load from xLabel.dll's folder
			string currentBuild = V_NAME;
			currentBuild.append("-");
			currentBuild.append(V_VERSION);
			currentBuild.append(".dll");

			GetModuleFileName(GetModuleHandle(currentBuild.c_str()), tmpSz, 4096);
			tmpSz[strlen(tmpSz) - currentBuild.length()] = 0;	// Remove the DLL name
			strcat(tmpSz, pluginName.c_str());
			m_PluginDesktopWatcher = LoadLibrary(tmpSz);

			if(m_PluginDesktopWatcher == NULL)
			{
				pluginName = "Plugin not found!\n\n";
				pluginName.append(tmpSz);
				pluginName.append("\n\nPlease check Path and PluginName!\nIf it still refuses to work, copy xdesktopwatcher.dll into your LitestepDir!");
				MessageBox(NULL, pluginName.c_str(), "xLabel Plugin Informations", MB_SETFOREGROUND);
			}
		}

		if (m_PluginDesktopWatcher != NULL)
		{
			pStart = (START)GetProcAddress(m_PluginDesktopWatcher, "OnStart");
			pStop = (STOP)GetProcAddress(m_PluginDesktopWatcher, "OnStop");
			
			if (pStart == NULL || pStop == NULL)
				MessageBox(NULL, "Couldn't initialise PluginFunctions!\nMaybe you use an older xdesktopwatcher.dll version, than xlabel.dll version!", "xLabel Plugin Informations", MB_SETFOREGROUND);
			else
			{
				char path[MAX_PATH];
				
				VarExpansion( path, "$desktopdir$" );
				pStart( path );
				VarExpansion( path, "$commondesktopdir$" );
				pStart( path );
			}
		}
	}

	return 0;
}

extern HDC hdcDesktop;
extern HBITMAP hbmDesktop;

void quitModule(HINSTANCE hInstance)
{
	RemoveBangCommand("!LabelAlwaysOnTop");
	RemoveBangCommand("!LabelClipboardCopy");
	RemoveBangCommand("!LabelClipboardPaste");
	RemoveBangCommand("!LabelCreate");
	RemoveBangCommand("!LabelDebug");
	RemoveBangCommand("!LabelDestroy");
	RemoveBangCommand("!LabelHide");
	RemoveBangCommand("!LabelLSBoxHook");
	RemoveBangCommand("!LabelMove");
	RemoveBangCommand("!LabelNext");
	RemoveBangCommand("!LabelPinToDesktop");
	RemoveBangCommand("!LabelPrevious");
	RemoveBangCommand("!LabelRefresh");
	RemoveBangCommand("!LabelReposition");
	RemoveBangCommand("!LabelResize");
	RemoveBangCommand("!LabelScroll");
	RemoveBangCommand("!LabelSetFontColor");
	RemoveBangCommand("!LabelSetFontShadowColor");
	RemoveBangCommand("!LabelSetText");
	RemoveBangCommand("!LabelShow");
	RemoveBangCommand("!LabelShowHide");
	RemoveBangCommand("!LabelToggleAlwaysOnTop");
	RemoveBangCommand("!LabelToggle");
	RemoveBangCommand("!LabelUpdate");
	if(IsOS(OS_2KXP))
		RemoveBangCommand("!LabelSetAlpha");
	RemoveBangCommand("!LabelToggleGhosted");

	RemoveBangCommand("!LabelmzscriptVarCopy");
	RemoveBangCommand("!LabelPosmzscriptVarCopy");
	RemoveBangCommand("!LabelSizemzscriptVarCopy");

	RemoveBangCommand("!LabelSetAnimation");

	RemoveBangCommand("!LabelShrink");
	RemoveBangCommand("!LabelGrow");
	RemoveBangCommand("!LabelMoveBy");

	RemoveBangCommand("!ParseEvars");
	RemoveBangCommand("!SetEvar");

	//Desktop
	RemoveBangCommand("!LabelDesktopCreate");
	RemoveBangCommand("!LabelDesktopSave");

	//TextEdit 2.52
	if (usetextedit)
	{
		RemoveBangCommand("!textappend");
		RemoveBangCommand("!textdel");
		RemoveBangCommand("!textreplace");
	}

	//Shortcut2
	if (useshortcut2)
	{
		RemoveBangCommand("!ShortcutGroupHide");
		RemoveBangCommand("!ShortcutGroupShow");
		RemoveBangCommand("!ShortcutGroupToggle");
		RemoveBangCommand("!ShortcutGroupOnTop");
		RemoveBangCommand("!ShortcutGroupOnBottom");
		RemoveBangCommand("!ShortcutGroupOnTopToggle");
	}

	LabelListIterator lit;

	for(lit = labelList.begin(); lit != labelList.end(); lit++)
		delete *lit;

	labelList.clear();

	for(lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
		delete *lit;

	desktoplabelList.clear();

	if (useshortcut2)
	{
		for(lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
			delete *lit;

		shortcutlabelList.clear();
	}

	if (useautodesktop)
	{
		if (!labelDesktopItems.empty())
		{	
			int counter = 1;

			//Clear all *xlabeldesktopitem lines
			Bang = ("@" + desktopSavePath + "@ @\\*[xX][lL][aA][bB][eE][lL][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@");
			BangTextEdittextDelete(NULL, Bang.c_str());

			for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
			{
				char buffer[32];
				strcpy(buffer, "xldi");
				char counterbuffer[16];
				itoa( counter, counterbuffer, 10 );
				if (counter < 10)
					strcat(buffer, "0");
				strcat(buffer, counterbuffer);

				char name[64], data[MAX_LINE_LENGTH];
				char *linebuffers[] = {name};

				LCTokenize((*sit).c_str(), linebuffers, 1, data);

				Bang = ("@" + desktopSavePath + "@ @*xlabeldesktopitem " + buffer + " " + data + "@");
				BangTextEdittextAppend(NULL, Bang.c_str());

				counter++;
			}
		}

		if (pStop != NULL)
			pStop();

		if (m_PluginDesktopWatcher != NULL)
		{
			FreeLibrary(m_PluginDesktopWatcher);
			m_PluginDesktopWatcher = NULL;
		}
	}

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xLabel", hInstance);
	UnregisterClass("xLabelMessageHandler", hInstance);

	delete systemInfo;
	delete defaultSettings;

	hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
	DeleteDC(hdcDesktop);
	DeleteObject(hbmDesktop);
}

Label *lookupLabel(const string &name, LabelList searchList)
{
	LabelListIterator it;

	for(it = searchList.begin(); it != searchList.end(); it++)
	{
		if(stricmp(name.c_str(), (*it)->getName().c_str()) == 0)
			return *it;
	}

	//LSLogPrintf(LOG_WARNING, szLogName, "Label \"%s\" not found!\nCheck Bang Command.", name.c_str());
	return 0;
}

BOOL APIENTRY DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID pvReserved)
{
	if(dwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hInstance);
	}

	return TRUE;
}

//Read all required Files/Folder Paths (in StringList)
StringList getActualDesktopItemPaths(bool defaultDesktop, string searchPath)
{
	StringList ActualDesktopItemPaths;

	string wildcardPath = "";
	string itemName = "";
	
	char tmp[MAX_PATH];
	string tmpHelper = "";

	HANDLE hFindHandle;
	WIN32_FIND_DATA FindData;

	if (defaultDesktop)
	{
		//Create 4 Main ShellFolders: Workspace, My Documents, NetworkHood, Trash 
		ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}");
		ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}");
		ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}");
		ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{645FF040-5081-101B-9F08-00AA002F954E}");

		VarExpansion( tmp, "$commondesktopdir$" );
		searchPath = tmp;
		wildcardPath = (searchPath + "*.*");

		hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;

			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}

		VarExpansion( tmp, "$desktopdir$" );
		searchPath = tmp;
		wildcardPath = (searchPath + "*.*");

		hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;
					
			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}
	}
	else
	{
		//Check for ending BackSlash
		tmpHelper = searchPath.substr(searchPath.length()-1, searchPath.length());

		if (tmpHelper != "\\")
			wildcardPath = (searchPath + "\\*.*");
		else
			wildcardPath = (searchPath + "*.*");

		hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;
					
			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}
	}

	FindClose(hFindHandle);

	return ActualDesktopItemPaths;
}

//Write all Items to File
void DesktopUpdate(int updateMode, string updatedFile, StringList ActualDesktopItemPaths)
{
	int counter = 1;

	//Saves already present Items, Only if:
	//	1. They exist
	//  2. It's wished
	//  3. AutoCreateDesktop isn't running
	if (!labelDesktopItems.empty() && updateMode == 1 && !useautodesktop)
	{
		for(StringListIterator it = labelDesktopItems.begin(); it != labelDesktopItems.end(); it++)
		{
			char buffer[32];
			strcpy(buffer, "xldi");
			char counterbuffer[16];
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			char name[64], data[MAX_LINE_LENGTH];
			char *linebuffers[] = {name};

			LCTokenize((*it).c_str(), linebuffers, 1, data);

			Bang = ("@" + desktopSavePath + "@ @*xlabeldesktopitem " + buffer + " " + data + "@");
			BangTextEdittextAppend(NULL, Bang.c_str());

			counter++;
		}
	}
	
	//Saves new Items, Only if:
	//	1. It's wished (Replace)
	//  2. It's wished (Append)
	//  3. None exists (Initial Creation required)
	//  AND IT'S NOT AN UPDATE
	if ( (updateMode == 0 || updateMode == 1 || labelDesktopItems.empty()) && updateMode != -1)
	{
		int tempcounter = 1;

		NameValuePair directionValues[] = {
		{ "left", 1 },
		{ "right", 2 },
		{ "up", 3 },
		{ "down", 4 },
		{ 0, 0 }
		};

		NameValuePair wrapdirectionValues[] = {
		{ "up", 1 },
		{ "down", 2 },
		{ "left", 3 },
		{ "right", 4 },
		{ 0, 0 }
		};

		int DIwidth = GetxRCDimen("xlabeldesktop", "ItemWidth", 100, GetSystemMetrics(SM_CXSCREEN));
		int DIheight = GetxRCDimen("xlabeldesktop", "ItemHeight", 75, GetSystemMetrics(SM_CYSCREEN));
		int xSpace = GetxRCInt("xlabeldesktop", "SpacingX", 0);
		int ySpace = GetxRCInt("xlabeldesktop", "SpacingY", 0);
				
		int direction = GetxRCNamedValue("xlabeldesktop", "Direction", directionValues, 2);
		int wrapDirection = GetxRCNamedValue("xlabeldesktop", "WrapDirection", wrapdirectionValues, 2);
		
		int wrapCount;
		if (wrapDirection == 1 || wrapDirection == 2)
			wrapCount = GetxRCInt("xlabeldesktop", "WrapCount", (GetSystemMetrics(SM_CXSCREEN)-xSpace)/(DIwidth+xSpace));
		else if (wrapDirection == 3 || wrapDirection == 4)
			wrapCount = GetxRCInt("xlabeldesktop", "WrapCount", (GetSystemMetrics(SM_CYSCREEN)-ySpace)/(DIheight+ySpace));

		int DIStartx = GetxRCCoord("xlabeldesktop", "X", ((GetSystemMetrics(SM_CXSCREEN)-(wrapCount*(DIwidth+xSpace)-xSpace))/2 < 0 ? 0 : (GetSystemMetrics(SM_CXSCREEN)-(wrapCount*(DIwidth+xSpace)-xSpace))/2), GetSystemMetrics(SM_CXSCREEN));
		int DIStarty = GetxRCCoord("xlabeldesktop", "Y", ((GetSystemMetrics(SM_CYSCREEN)-(wrapCount*(DIheight+ySpace)-ySpace))/2 < 50 ? 50 : (GetSystemMetrics(SM_CYSCREEN)-(wrapCount*(DIheight+ySpace)-ySpace))/2), GetSystemMetrics(SM_CYSCREEN));

		int tempx;
		int tempy;

		if (direction == 1)
			tempx = DIStartx+DIwidth+xSpace;
		else if (direction == 2)
			tempx = DIStartx-(DIwidth+xSpace);
		else if (direction == 3)
			tempy = DIStarty+DIheight+ySpace;
		else if (direction == 4)
			tempy = DIStarty-(DIheight+ySpace);

		if (direction == 1 || direction == 2)
			tempy = DIStarty;
		else if (direction == 3 || direction == 4)
			tempx = DIStartx;

		for(StringListIterator it = ActualDesktopItemPaths.begin(); it != ActualDesktopItemPaths.end(); it++)
		{
			char buffer[32];
			strcpy(buffer, "xldi");
			char counterbuffer[16];
			char xbuffer[16];
			char ybuffer[16];
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			//Sorting/Positioning due to settings
			div_t div_result = div( tempcounter, wrapCount );

			if (direction == 1 || direction == 2)
			{
				if (wrapDirection > 2)
					wrapDirection = 2;
						
				if (div_result.quot > 0 && div_result.rem == 1)
				{
					tempx = DIStartx;
					if (wrapDirection == 1)
						tempy = DIStarty-(div_result.quot*(DIheight+ySpace));
					else if (wrapDirection == 2)
						tempy = DIStarty+div_result.quot*(DIheight+ySpace);
				}
				else
				{
					if (direction == 1)
						tempx = tempx-(DIwidth+xSpace);
					else if (direction == 2)
						tempx = tempx+DIwidth+xSpace;
				}
			}
			if (direction == 3 || direction == 4)
			{
				if (wrapDirection < 3)
					wrapDirection = 4;
					
				if (div_result.quot > 0 && div_result.rem == 1)
				{
					tempy = DIStarty;
					if (wrapDirection == 3)
						tempx = DIStartx-(div_result.quot*(DIwidth+xSpace));
					else if (wrapDirection == 4)
						tempx = DIStartx+div_result.quot*(DIwidth+xSpace);
				}
				else
				{
					if (direction == 3)
						tempy = tempy-(DIheight+ySpace);
					else if (direction == 4)
						tempy = tempy+DIheight+ySpace;
				}
			}

			itoa( tempx, xbuffer, 10 );
			itoa( tempy, ybuffer, 10 );

			if (!useautodesktop)
			{
				Bang = ("@" + desktopSavePath + "@ @*xlabeldesktopitem " + buffer + " " + xbuffer + " " + ybuffer + " \"" + (*it) + "\"@");
				BangTextEdittextAppend(NULL, Bang.c_str());
			}

			Label *label = new Label(buffer);

			desktoplabelList.insert(desktoplabelList.end(), label);

			Bang = "dummy ";
			Bang.append( xbuffer );
			Bang.append( " ");
			Bang.append( ybuffer );
			Bang.append( " \"" + (*it) + "\"");

			label->desktopload(hInstance, Bang.c_str());

			labelDesktopItems.insert(labelDesktopItems.end(), Bang);

			counter++;
			tempcounter++;
		}
	}
	//Update & Saves new Items, Only if:
	//	1. Some have changed
	else if (updateMode == -1)
	{
		for (StringListIterator sit = ActualDesktopItemPaths.begin(); sit != ActualDesktopItemPaths.end(); sit++)
		{
			bool found = false;

			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				if (stricmp((*sit).c_str(), (*lit)->iconCommand.c_str()) == NULL)
				{
					found = true;
					break;
				}
			}

			if (found)
			{
				int left = (*lit)->getX();
				int top = (*lit)->getY();

				if (left < 0)
					left = 0;
				if (top < 0)
					top = 0;
							
				string position = ( IntToString(left) + " " + IntToString(top) );

				Bang = ( "dummy " + position + " \"" + (*lit)->iconCommand + "\"");

				labelDesktopItems.insert(labelDesktopItems.end(), Bang);

				(*lit)->hide();
				desktoplabelList.remove((*lit));
				delete (*lit);
			}
			else
			{
				Bang = ( "dummy 100 100 \"" + (*sit) + "\"" );
			
				labelDesktopItems.insert(labelDesktopItems.end(), Bang);
			}
		}
		for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
		{
			(*lit)->hide();
			desktoplabelList.remove((*lit));
			delete (*lit);
		}
		
		int counter = 1;
		for(sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
		{
			char buffer[32];
			strcpy(buffer, "xldi");
			char counterbuffer[16];
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			char name[64], data[MAX_LINE_LENGTH];
			char *linebuffers[] = {name};

			LCTokenize((*sit).c_str(), linebuffers, 1, data);

			Label *label = new Label(name);

			desktoplabelList.insert(desktoplabelList.end(), label);

			label->desktopload(hInstance, (*sit));

			if (!useautodesktop)
			{
				Bang = ("@" + desktopSavePath + "@ @*xlabeldesktopitem " + buffer + " " + data + "@");
				BangTextEdittextAppend(NULL, Bang.c_str());
			}
			counter++;
		}
	}
	//Update & Saves single changed Items, Only if:
	//	1. they have changed
	//	1. AutoCreateDesktop is ON
	else if (updateMode >= 20 && useautodesktop)
	{
		//Add Files
		if (updateMode == 21)
		{
			Bang = ( "dummy 100 100 \"" + updatedFile + "\"" );
		
			labelDesktopItems.insert(labelDesktopItems.end(), Bang);

			Label *label = new Label("dummy");

			desktoplabelList.insert(desktoplabelList.end(), label);

			label->desktopload(hInstance, Bang.c_str());
		}
		//Remove Files
		else if (updateMode == 22)
		{
			bool found = false;
			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				if (stricmp((*lit)->iconCommand.c_str(), updatedFile.c_str()) == NULL)
				{
					found = true;
					break;
				}
			}
			
			if (found)
			{
				char xname[32], xx[16], xy[16], xpath[MAX_PATH];
				char *buffers[] = {xname, xx, xy, xpath};

				for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
				{
					LCTokenize((*sit).c_str(), buffers, 4, NULL);

					if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
					{
						labelDesktopItems.remove( *sit );
						break;
					}
				}

				(*lit)->hide();
				desktoplabelList.remove((*lit));
				delete (*lit);
			}
		}
		//Rename Files
		else if (updateMode == 23)
		{
			HANDLE hFindHandle;
			WIN32_FIND_DATA FindData;

			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				if (strstr((*lit)->iconCommand.c_str(), "::") == NULL)
				{
					hFindHandle = FindFirstFile((*lit)->iconCommand.c_str(), &FindData);

					if(INVALID_HANDLE_VALUE == hFindHandle)
					{
						char xname[32], xx[16], xy[16], xpath[MAX_PATH];
						char *buffers[] = {xname, xx, xy, xpath};

						for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
						{
							LCTokenize((*sit).c_str(), buffers, 4, NULL);

							if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
							{
								labelDesktopItems.remove( *sit );
								break;
							}
						}

						string positions = ( IntToString((*lit)->getX()) + " " + IntToString((*lit)->getY()) );

						(*lit)->hide();
						delete *lit;
						desktoplabelList.remove( *lit );

						Bang = ( "dummy " + positions + " \"" + updatedFile + "\"" );

						//New Named Label
						labelDesktopItems.insert(labelDesktopItems.end(), Bang);

						Label *label = new Label("dummy");

						desktoplabelList.insert(desktoplabelList.end(), label);

						label->desktopload(hInstance, Bang.c_str());

						break;
					}
				}
			}

			FindClose(hFindHandle);
		}
	}
}

// !LabelDesktopCreate <append>|<replace>|<update...> <path to lnkfolder>|<updatefile>
void BangDesktopCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	StringList labelDesktopItemPaths;

	char action[32], path[MAX_LINE_LENGTH];
	char *buffers[] = {action, path};

	LCTokenize(pszArgs, buffers, 2, NULL);
	
	string searchPath = path;

	//Quit here, if no path is specified
	if (searchPath.empty())
		return;

	//Initialize
	bool defaultDesktop = false;
	
	if ( stricmp(searchPath.c_str(), ".desktop") == 0 || stricmp(searchPath.c_str(), "deskto") == 0 || stricmp(searchPath.c_str(), "eskto") == 0 )
		defaultDesktop = true;
	
	//Start Configuration and Writing to File
	if (!desktopSavePath.empty())
	{
		int updateMode;
		if ( stricmp(action, "update") == NULL )
			updateMode = -1;
		else if ( stricmp(action, "updateadd") == NULL )
			updateMode = 21;
		else if ( stricmp(action, "updatedel") == NULL )
			updateMode = 22;
		else if ( stricmp(action, "updateren") == NULL )
			updateMode = 23;
		else if (stricmp(action, "append") == NULL)
			updateMode = 1;
		else 
			updateMode = 0;

		if (updateMode == -1 || updateMode == 0 || updateMode == 1)
		{
			//Get All Items actual items
			labelDesktopItemPaths = getActualDesktopItemPaths(defaultDesktop, searchPath);

			//Clear all *xlabeldesktopitem lines
			Bang = ("@" + desktopSavePath + "@ @\\*[xX][lL][aA][bB][eE][lL][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@");
			BangTextEdittextDelete(NULL, Bang.c_str());

			if (updateMode == -1 || updateMode == 0)
			{
				labelDesktopItems.clear();

				if (updateMode == 0)
				{
					Bang = "!labeldestroy xlabeldesktop";
					LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
				}
			}
		}

		//0 = Create New (Replace/Reposition), 1 = Append (New will be Positioned), 2+ = Update
		DesktopUpdate(updateMode, searchPath, labelDesktopItemPaths);
	}
	else
	{
		Bang = "!alert \"You haven't defined the setting xLabelDesktopConfigFile, which points to the file containing your *xLabelDesktopItem lines!\"";
		LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
	}
}

// !LabelDesktopSave [internal caller]
void BangDesktopSave(HWND hwndCaller, LPCSTR pszArgs)
{
	string internalcaller = pszArgs;

	if (!desktoplabelList.empty() && internalcaller.empty() && !useautodesktop)
	{
		for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
		{
			int left = (*lit)->getX();
			int top = (*lit)->getY();

			if (left < 0)
				left = 0;
			if (top < 0)
				top = 0;
						
			string position = ( IntToString(left) + " " + IntToString(top) );

			Bang = ("@" + (*lit)->configSavePath + "@ @[		]*(\\*[xX][lL][aA][bB][eE][lL][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]).*(" + (*lit)->getName() + ")@ @\\1 \\2 " + position + " \"" + (*lit)->iconCommand + "\"@");
			if (!(*lit)->configSavePath.empty())
				BangTextEdittextReplace(NULL, Bang.c_str());
			else
			{
				Bang = "!alert \"You haven't defined the setting xLabelDesktopConfigFile, which points to the file containing your *xLabelDesktopItem lines!\"";
				LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
			}
		}
		//Stop if complete Desktop was saved
		return;
	}
	else
	{
		for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
		{
			if (stricmp((*lit)->iconCommand.c_str(), internalcaller.c_str()) == NULL)
			{
				int left = (*lit)->getX();
				int top = (*lit)->getY();

				if (left < 0)
					left = 0;
				if (top < 0)
					top = 0;
						
				string position = ( IntToString(left) + " " + IntToString(top) );

				//Only Track change and do not save, if AutoCreate Desktop
				if (useautodesktop)
				{
					char xname[32], xx[16], xy[16], xpath[MAX_PATH];
					char *buffers[] = {xname, xx, xy, xpath};

					for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
					{
						LCTokenize((*sit).c_str(), buffers, 4, NULL);

						if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
						{
							Bang = ( (*lit)->getName() + " " + position + " \"" + (*lit)->iconCommand + "\"" );
							
							*sit = Bang;
							
							break;
						}
					}
				}
				//Save immediately, if no AutoCreate Desktop
				else
				{
					Bang = ("@" + (*lit)->configSavePath + "@ @[		]*(\\*[xX][lL][aA][bB][eE][lL][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]).*(" + (*lit)->getName() + ")@ @\\1 \\2 " + position + " \"" + (*lit)->iconCommand + "\"@");
					if (!(*lit)->configSavePath.empty())
						BangTextEdittextReplace(NULL, Bang.c_str());
				}

				return;
			}
		}
	}
	//No Complete Desktop Save and No Single DesktopItem Save
	//Try Shortcuts ;) 
	Label *label = lookupLabel(internalcaller.c_str(), shortcutlabelList);

	if(label)
	{
		int left = label->getX();
		int top = label->getY();

		if (left < 0)
			left = 0;
		if (top < 0)
			top = 0;
					
		string position = ( IntToString(left) + " " + IntToString(top) );
		
		Bang = ("@" + label->configSavePath + "@ @[		]*(\\*[sS][hH][oO][rR][tT][cC][uU][tT])([ 	]*)(\\=" + label->getName() + "\\=)([ 	]*)([^ 	]+)([ 	]*)([^ 	]+)([ 	]*)(.*)@ @\\1 \\3 " + position + " \\9@");
		if (!label->configSavePath.empty())
			BangTextEdittextReplace(NULL, Bang.c_str());
		else
		{
			Bang = "!alert \"You haven't defined the setting ShortcutConfigFile, which points to the file containing your *shortcut lines!\"";
			LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
		}
	}
}

// !Labelmzscriptvarcopy <label> <mzscript var>
void BangmzscriptVarCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], var[64];
	char *buffers[] = {name, var};

	if(LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->mzscriptvarcopy(var);
	}
}

// !LabelPosmzscriptvarcopy <label> <mzscript var xpos> <mzscript var ypos>
void BangPosmzscriptVarCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], varx[64], vary[64];
	char *buffers[] = {name, varx, vary};

	if(LCTokenize(pszArgs, buffers, 3, NULL) == 3)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->posmzscriptvarcopy(varx, vary);
	}
}

// !LabelSizemzscriptvarcopy <label> <mzscript var cx> <mzscript var cy>
void BangSizemzscriptVarCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], varcx[64], varcy[64];
	char *buffers[] = {name, varcx, varcy};

	if(LCTokenize(pszArgs, buffers, 3, NULL) == 3)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->sizemzscriptvarcopy(varcx, varcy);
	}
}

// !LabelSetAnimation <label> <on | off> [<loops>]
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16], loops[16];
	char *buffers[] = {name, option, loops};

	int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);
	if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (numTokens == 2)
			{
				if(stricmp(option, "off") == 0)
					label->setAnimation(false);
				else if(stricmp(option, "on") == 0)
					label->setAnimation(true);
			}
			else if  (numTokens == 3)
			{
				if(stricmp(option, "off") == 0)
					label->setAnimation(false);
				else if(stricmp(option, "on") == 0)
					label->setAnimation(true, atoi(loops));
			}
		}
	}
}

// !LabelAlwaysOnTop <label>
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->setAlwaysOnTop(true);
	else
	{
		LabelListIterator i;
		for( i = labelList.begin(); i != labelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->setAlwaysOnTop(true);
		}
		for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->setAlwaysOnTop(true);
		}
	}
}

// !LabelClipboardCopy <label> <prefix>
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardCopy((numTokens >= 2) ? prefix : "");
	}
}

// !LabelClipboardPaste <label> <prefix>
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardPaste((numTokens >= 2) ? prefix : "");
	}
}

// !LabelCreate <label>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = new Label(pszArgs);
	label->load(hInstance);
	labelList.insert(labelList.end(), label);
}

// !LabelDebug
void BangDebug(HWND hwndCaller, LPCSTR pszArgs)
{
	LabelListIterator it;
	string message = "Normal Labels: (*label)\r\n";

	for(it = labelList.begin(); it != labelList.end(); it++)
	{
		message.append((*it)->getName());
		message.append(", ");
	}

	message.append("\r\n\r\nDesktop Icons: (*xlabeldesktopitem)\r\n");

	for(it = desktoplabelList.begin(); it != desktoplabelList.end(); it++)
	{
		message.append((*it)->getName());
		message.append(", ");
	}

	message.append("\r\n\r\nShortcut Lines: (*shortcut)\r\n");

	for(it = shortcutlabelList.begin(); it != shortcutlabelList.end(); it++)
	{
		message.append((*it)->getName());
		message.append("\r\n");
	}

	MessageBox(NULL, message.c_str(), "xLabel Debug Informations", MB_SETFOREGROUND);
}

// !LabelDestroy <label>
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{
	if (stricmp(pszArgs, "xlabeldesktop") == 0)
	{
		for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			delete *lit;

		desktoplabelList.clear();
	}
	else
	{
		Label *label = lookupLabel(pszArgs);
		if(label)
		{
			label->hide();
			labelList.remove(label);
			delete label;
		}
		else
		{
			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
				{
					(*lit)->hide();
					labelList.remove((*lit));
					delete (*lit);
				}
			}
		}
	}
}

// !LabelHide <label>
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
	if (stricmp(pszArgs, "xlabeldesktop") == 0)
	{
		for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			(*lit)->hide();
	}
	else
	{
		Label *label = lookupLabel(pszArgs);
		if(label)
			label->hide();
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++ )
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
					(*lit)->hide();
			}
			for( lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++ )
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
					(*lit)->hide();
			}
		}
	}
}

// !LabelLSBoxHook <labelname>
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], dummyx[32], dummyy[32], dummyn[32];
	char *buffers[] = {name, dummyx, dummyy, dummyn};

	int tokens = LCTokenize(pszArgs, buffers, 4, 0);

	const char *handle = strrchr(pszArgs,' ');

	HWND hwndBox = (HWND)atoi(handle+1);

	Label *label = lookupLabel(name);

	if(label)
	{
		label->setBox(hwndBox);
		label->update();		
	}
	else if(tokens <= 3)
	{
		label = new Label(name);
		labelList.insert(labelList.end(), label);
		label->load(hInstance, hwndBox);
	}
	else
	{
		label = new Label(name);
		shortcutlabelList.insert(shortcutlabelList.end(), label);
		label->shortcutload(hInstance, pszArgs, 0, 0, hwndBox);
	}
}

// !LabelMove <label> <x> <y> [<steps> <time>]
void BangMove(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if(LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, GetSystemMetrics(SM_CXSCREEN));
				y = xParseCoord(B, 0, GetSystemMetrics(SM_CYSCREEN));
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right-rc.left);
				y = xParseCoord(B, 0, rc.bottom-rc.top);
			}
			int step = atoi( C );
			int time = atoi( D );

			label->move(x, y, step, time);
		}
	}
}

// !LabelMoveBy <label> <x> <y>
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16];
	char *buffers[] = {name, A, B};

	if(LCTokenize(pszArgs, buffers, 3, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int x = atoi( A );
		int y = atoi( B );

		if(label)
			label->move(x, y, 0, 0, true);
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->move(x, y, 0, 0, true);
			}
		}
	}
}

// !LabelNext <label>
void BangNext(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->next();
}

// !LabelPinToDesktop <label>
void BangPinToDesktop(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->setAlwaysOnTop(false);
	else
	{
		LabelListIterator i;
		for( i = labelList.begin(); i != labelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->setAlwaysOnTop(false);
		}
		for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->setAlwaysOnTop(false);
		}
	}
}

// !LabelPrevious <label>
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->previous();
}

// !LabelRefresh <label>
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], setting[256], value[256];
	char *buffers[] = {name, setting, value};

	int tokens = LCTokenize(pszArgs, buffers, 3, NULL);

	Label *label = lookupLabel(name);

	if(label)
	{
		if (tokens == 1)
		{
			label->reconfigure();
			label->repaint(true);
		}

		if (tokens == 3)
		{
			Bang = "\"";
			Bang.append( value );
			Bang.append("\"");

			PrefixedLSSetVariable((string) setting, "", Bang);
			label->reconfigure();
			label->repaint(true);
		}
	}
	else
	{
		LabelListIterator lit;
		for( lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
			{
				if (tokens == 1)
				{
					label->reconfigure();
					label->repaint(true);
				}

				if (tokens == 3)
				{
					string worker = setting;

					if (strstr(setting, "*") != NULL)
						worker.replace(worker.find("*"), 1, (*lit)->getName().c_str(), (*lit)->getName().length());
					
					Bang = "\"";
					Bang.append( value );
					Bang.append("\"");

					PrefixedLSSetVariable(worker, "", Bang);
					(*lit)->reconfigure();
					(*lit)->repaint(true);
				}
			}
		}
	}
}

// !LabelReposition <label> <x> <y> <width> <height> [<steps>] [<time>]
void BangReposition(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if(LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, GetSystemMetrics(SM_CXSCREEN));
				y = xParseCoord(B, 0, GetSystemMetrics(SM_CYSCREEN));
				w = xParseDimen(C, 0, GetSystemMetrics(SM_CXSCREEN));
				h = xParseDimen(D, 0, GetSystemMetrics(SM_CYSCREEN));
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right-rc.left);
				y = xParseCoord(B, 0, rc.bottom-rc.top);
				w = xParseDimen(C, 0, rc.right-rc.left);
				h = xParseDimen(D, 0, rc.bottom-rc.top);
			}
			int step = atoi( E );
			int time = atoi( F );

			label->reposition(x, y, w, h, step, time);
		}
	}
}

// !LabelResize <label> <width> <height> [<steps> <time>]
void BangResize(HWND hwndCaller, LPCSTR pszArgs)
{
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if(LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (label->getBox() == NULL)
			{
				w = xParseDimen(A, 0, GetSystemMetrics(SM_CXSCREEN));
				h = xParseDimen(B, 0, GetSystemMetrics(SM_CYSCREEN));
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				w = xParseDimen(A, 0, rc.right-rc.left);
				h = xParseDimen(B, 0, rc.bottom-rc.top);
			}
			int step = atoi( C );
			int time = atoi( D );

			label->resize(w, h, step, time);
		}
	}
}

// !LabelShrink <label> <cx> <cy>
void BangShrink(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16];
	char *buffers[] = {name, A, B};

	if(LCTokenize(pszArgs, buffers, 3, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int cx = atoi( A );
		int cy = atoi( B );

		if(label)
			label->resize(cx, cy, 0, 0, -1);
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->resize(cx, cy, 0, 0, -1);
			}
		}
	}
}

// !LabelGrow <label> <cx> <cy>
void BangGrow(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16];
	char *buffers[] = {name, A, B};

	if(LCTokenize(pszArgs, buffers, 3, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int cx = atoi( A );
		int cy = atoi( B );

		if(label)
			label->resize(cx, cy, 0, 0, 1);
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->resize(cx, cy, 0, 0, 1);
			}
		}
	}
}

// !LabelScroll <label> <'on', 'off', 'toggle', 'freeze', limit> <scrollMode(1-4)>
void BangScroll(HWND hwndCaller, LPCSTR pszArgs)
{
	// FIXME: This should be split into multiple bang commands
	char name[64], option[16], mode[16];
	char *buffers[] = {name, option, mode};

	if(LCTokenize(pszArgs, buffers, 3, NULL) >= 2)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if(stricmp(option, "off") == 0)
			{
				label->setScrolling(0);
				PrefixedLSSetVariable( name, "Scroll", false );
			}
			else if(stricmp(option, "on") == 0)
			{
				int scrollmode = atoi(mode);
				if (scrollmode > 0 && scrollmode <= 4)
					label->setScrolling(scrollmode);
				else
					label->setScrolling(label->scrollBackup);
			}
			else if(stricmp(option, "toggle") == 0)
			{
				if (label->getScrolling() > 0)
				{
					label->setScrolling(0);
					PrefixedLSSetVariable( name, "Scroll", false );
				}
				else
				{
					int scrollmode = atoi(mode);
					if (scrollmode > 0 && scrollmode <= 4)
						label->setScrolling(scrollmode);
					else
						label->setScrolling(label->scrollBackup);
				}
			}
			else if(stricmp(option, "freeze") == 0)
			{
				if (label->frozen)
					label->frozen = false;
				else
					label->frozen = true;
			}
			else
			{
				int limit = atoi(option);
				if(limit >= 0) label->setScrollLimit(limit);
			}
		}
	}
}

// !LabelSetFontColor <label> <red> <green> <blue> OR !LabelSetFontColor <label> <hexcolor>
void BangSetFontColor(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16];
	char *buffers[] = {name, A, B, C};
	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);
	int color;

	if(numTokens >= 2)
	{
		if (stricmp(name, "xlabeldesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			{
				if(numTokens >= 4)
				{
					int r = atoi(A);
					int g = atoi(B);
					int b = atoi(C);

					color = RGB(r, g, b);
				}
				else
				{
					color = strtol(A, NULL, 16);
					color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
				}

				(*i)->getFont()->setColor(color);
				(*i)->repaint();
			}
		}
		else
		{
			Label *label = lookupLabel(name);

			if (label)
			{
				if(numTokens >= 4)
				{
					int r = atoi(A);
					int g = atoi(B);
					int b = atoi(C);

					color = RGB(r, g, b);

					string setcol = A;
					setcol.append( " " );
					setcol.append( B );
					setcol.append( " " );
					setcol.append( C );
					PrefixedLSSetVariable( name, "FontColor", setcol );
				}
				else
				{
					color = strtol(A, NULL, 16);
					color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

					PrefixedLSSetVariable( name, "FontColor", (string) A );
				}

				label->getFont()->setColor(color);
				label->repaint();
			}
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
					{
						if(numTokens >= 4)
						{
							int r = atoi(A);
							int g = atoi(B);
							int b = atoi(C);

							color = RGB(r, g, b);

							string setcol = A;
							setcol.append( " " );
							setcol.append( B );
							setcol.append( " " );
							setcol.append( C );
							PrefixedLSSetVariable( name, "FontColor", setcol );
						}
						else
						{
							color = strtol(A, NULL, 16);
							color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

							PrefixedLSSetVariable( name, "FontColor", (string) A );
						}

						(*i)->getFont()->setColor(color);
						(*i)->repaint();
					}
				}
			}
		}
	}
}

// !LabelSetFontShadowColor <label> <red> <green> <blue> OR !LabelSetFontColor <label> <hexcolor>
void BangSetFontShadowColor(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16];
	char *buffers[] = {name, A, B, C};
	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);
	int color;

	if (numTokens == 1)
	{
		if (stricmp(name, "xlabeldesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			{
				(*i)->getFont()->setShadow(false);
				(*i)->repaint();
				if ((*i)->getTransparencyMode() == 1)
					(*i)->updateRegion();
			}
		}
		else
		{
			Label *label = lookupLabel(name);

			if (label)
			{
				label->getFont()->setShadow(false);
				label->repaint();
				if (label->getTransparencyMode() == 1)
					label->updateRegion();

				PrefixedLSSetVariable( name, "FontShadow", "false" );
			}
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
					{
						(*i)->getFont()->setShadow(false);
						(*i)->repaint();
						if ((*i)->getTransparencyMode() == 1)
							(*i)->updateRegion();

						PrefixedLSSetVariable( name, "FontShadow", "false" );
					}
				}
			}
		}
	}
	else if(numTokens >= 2)
	{
		if (stricmp(name, "xlabeldesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			{
				if(numTokens >= 4)
				{
					int r = atoi(A);
					int g = atoi(B);
					int b = atoi(C);

					color = RGB(r, g, b);
				}
				else
				{
					color = strtol(A, NULL, 16);
					color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
				}

				(*i)->getFont()->setColor(color, true);
				(*i)->getFont()->setShadow(true);
				(*i)->repaint();
				if ((*i)->getTransparencyMode() == 1)
					(*i)->updateRegion();
			}
		}
		else
		{
			Label *label = lookupLabel(name);

			if (label)
			{
				if(numTokens >= 4)
				{
					int r = atoi(A);
					int g = atoi(B);
					int b = atoi(C);

					color = RGB(r, g, b);

					string setcol = A;
					setcol.append( " " );
					setcol.append( B );
					setcol.append( " " );
					setcol.append( C );
					PrefixedLSSetVariable( name, "FontShadowColor", setcol );
				}
				else
				{
					color = strtol(A, NULL, 16);
					color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

					PrefixedLSSetVariable( name, "FontShadowColor", (string) A );
				}

				label->getFont()->setColor(color, true);
				label->getFont()->setShadow(true);
				label->repaint();
				if (label->getTransparencyMode() == 1)
					label->updateRegion();

				PrefixedLSSetVariable( name, "FontShadow", "true" );
			}
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
					{
						if(numTokens >= 4)
						{
							int r = atoi(A);
							int g = atoi(B);
							int b = atoi(C);

							color = RGB(r, g, b);

							string setcol = A;
							setcol.append( " " );
							setcol.append( B );
							setcol.append( " " );
							setcol.append( C );
							PrefixedLSSetVariable( name, "FontShadowColor", setcol );
						}
						else
						{
							color = strtol(A, NULL, 16);
							color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

							PrefixedLSSetVariable( name, "FontShadowColor", (string) A );
						}

						(*i)->getFont()->setColor(color, true);
						(*i)->getFont()->setShadow(true);
						(*i)->repaint();
						if ((*i)->getTransparencyMode() == 1)
							(*i)->updateRegion();

						PrefixedLSSetVariable( name, "FontShadow", "true" );
					}
				}
			}
		}
	}
}

// !LabelSetText <label> <text>
void BangSetText(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], text[MAX_LINE_LENGTH];
	char *buffers[] = {name, text};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->setText((numTokens >= 2) ? text : "");
	}
}

// !LabelShow <label>
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
	if (stricmp(pszArgs, "xlabeldesktop") == 0)
	{
		for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			(*i)->show();
	}
	else
	{
		Label *label = lookupLabel(pszArgs);
		if(label)
			label->show();
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->show();
			}
			for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->show();
			}
		}
	}
}

// !LabelShowHide <label> <timeout>
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[64];
	char *buffers[] = {name, A};

	if(LCTokenize(pszArgs, buffers, 2, NULL) >= 2)
	{
		int timeout = atoi(A);

		if (stricmp(name, "xlabeldesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				(*i)->showHide(timeout);
		}
		else
		{
			Label *label = lookupLabel(name);

			if(label)
				label->showHide(timeout);
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
						(*i)->showHide(timeout);
				}
				for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
						(*i)->showHide(timeout);
				}
			}
		}
	}
}

// !LabelToggleAlwaysOnTop <label>
void BangToggleAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->setAlwaysOnTop(!label->isAlwaysOnTop());
	else
	{
		LabelListIterator i;
		for( i = labelList.begin(); i != labelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->setAlwaysOnTop(!(*i)->isAlwaysOnTop());
		}
		for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->setAlwaysOnTop(!(*i)->isAlwaysOnTop());
		}
	}
}

// !LabelToggle <label>
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	if (stricmp(pszArgs, "xlabeldesktop") == 0)
	{
		for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
		{
			if((*i)->isVisible())
				(*i)->hide();
			else
				(*i)->show();
		}
	}
	else
	{
		Label *label = lookupLabel(pszArgs);
		if(label)
		{
			if(label->isVisible())
				label->hide();
			else
				label->show();
		}
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				{
					if((*i)->isVisible())
						(*i)->hide();
					else
						(*i)->show();
				}
			}
			for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				{
					if((*i)->isVisible())
						(*i)->hide();
					else
						(*i)->show();
				}
			}
		}
	}
}

// !LabelUpdate <label>
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs)
{
	if (stricmp(pszArgs, "xlabeldesktop") == 0)
	{
		for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			(*i)->update();
	}
	else
	{
		Label *label = lookupLabel(pszArgs);
		if(label)
			label->update();
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->update();
			}
			for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->update();
			}
		}
	}
}

// !LabelSetAlpha <label> <value>
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], factor[16];
	char *buffers[] = {name, factor};

	if(LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{
		if (stricmp(name, "xlabeldesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				(*i)->setAlpha(atoi( factor ));
		}
		else
		{
			Label *label = lookupLabel(name);
			if(label)
				label->setAlpha(atoi( factor ));
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
						(*i)->setAlpha(atoi( factor ));
				}
				for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
						(*i)->setAlpha(atoi( factor ));
				}
			}
		}
	}
}

// !LabelToggleGhosted <label>
void BangToggleGhosted(HWND hwndCaller, LPCSTR pszArgs)
{
	if (stricmp(pszArgs, "xlabeldesktop") == 0)
	{
		for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			(*i)->setGhosted(!(*i)->isGhosted());
	}
	else
	{
		Label *label = lookupLabel(pszArgs);
		if(label)
			label->setGhosted(!label->isGhosted());
		else
		{
			LabelListIterator i;
			for( i = labelList.begin(); i != labelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->setGhosted(!(*i)->isGhosted());
			}
			for( i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
			{
				if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
					(*i)->setGhosted(!(*i)->isGhosted());
			}
		}
	}
}

//TEXTEDIT 2.52 SECTION:
//=========================================================
// getToken grabs a token and clobbers info that is parsed
//=========================================================
void getToken(char line[4096], char temp[4096], bool escape_codes /*= true*/)
{
    int offset = 0, i;
	bool endquote = false;
    char ENCAP = '@';

	while (line[offset] == ' ')
	{
		offset++;
	}

	if (line[offset] == ENCAP)
	{
		offset++;
		endquote = true;
	}

	for (i = 0; i+offset < int(strlen(line)); ++i)
	{
		// Copy else if block to add more escape codes.
		if (escape_codes && (line[i+offset] == '\\' && int(strlen(line)) != i+offset+1))
		{
		    if (line[i+offset+1] == ENCAP || line[i+offset+1] == '\\')
			{
				offset++;
				temp[i] = line[i+offset];
				continue;
			}
			// Bang escape code below. Present in 2.0 from Tony Chang
			else if (line[i+offset+1] == '^')
			{
				offset++;
				temp[i] = '!';
				continue;
			}
			// EVar escape code below. Present in 2.0 from Tony Chang
			else if (line[i+offset+1] == '#')
			{
				offset++;
				temp[i] = '$';
				continue;
			}
			// Comment escape code below. Added in 2.1l by leaf r&d
			else if (line[i+offset+1] == '~')
			{
				offset++;
				temp[i] = ';';
				continue;
			}
			// Quotation mark escape code below. Added in 2.2l by leaf r&d
			else if (line[i+offset+1] == '=')
			{
				offset++;
                temp[i] = '"';
                continue;
            }
            // Clipboard escape code
            else if (line[i+offset+1] == 'c')
            {
				HGLOBAL hMem;
				LPVOID pvMem;

				if(IsClipboardFormatAvailable(CF_TEXT))
				{
					if(OpenClipboard(NULL))
					{
						hMem = GetClipboardData(CF_TEXT);

						if(hMem)
						{
							char pszBuffer[512];
							pvMem = GlobalLock(hMem);
							lstrcpyn(pszBuffer, (LPCTSTR) pvMem, 512);
							temp[i] = '\0';
							offset += min(strlen(pszBuffer), 512-strlen(temp));
							strncat(temp, pszBuffer, 512-strlen(temp));
							i = strlen(temp)+1;
							GlobalUnlock(hMem);
						}

						CloseClipboard();
					}
				}
            }
		}

 	    if (line[i+offset] == ENCAP && endquote)
	  	    break;
		else if (line[i+offset] == ' ' && !endquote)
			break;
		else
			temp[i] = line[i+offset];
	}
	temp[i] = '\0';

	if (i+offset+1 > int(strlen(line)))
	{
		line[0] = '\0';
	}
	else
	{
		strcpy(line, line+i+offset+1);
	}
}

//=========================================================
// Bang command handling
//=========================================================

void BangTextEdittextAppend(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[4096], str[4096];
	char rest[4096];
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, str, false);

	ofstream file;
	file.open(filename, /* ios::nocreate | */ ios::app);

	if (!file.fail())
	{
		file << str << endl;
	}

	file.close();
}

void BangTextEdittextDelete(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[4096], regex[4096], input[4096];
	char rest[4096];
	regexp *r;
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, regex, true);

	r = regcomp(regex);

	ifstream file;
	file.open(filename/*, ios::nocreate*/);        //removed because dev-cpp wont compile it... feel free to recompile to get it to work

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, 4096))
		{
			if (!regexec(r, input))
			{
				s = input;
				l.push_back(s);
			}
		}
		file.close();
		ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
}

void BangTextEdittextReplace(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[4096], regex[4096], input[4096], replace[4096];
	char rest[4096], buf[4096];
	regexp *r;
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, regex, true);
	getToken(rest, replace, true);

	r = regcomp(regex);

	ifstream file;
	file.open(filename/*, ios::nocreate*/);        //removed because dev-cpp wont compile it... feel free to recompile to get it to work

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, 4096))
		{
			if (regexec(r, input))
			{
				regsub(r, replace, buf);
				s = buf;
				l.push_back(buf);
			}
			else
			{
				s = input;
				l.push_back(input);
			}
		}
		file.close();
		ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
}

//Shortcut 2
void BangShortcutHide(HWND hwndCaller, LPCSTR pszArgs)
{	
	char group[4096];
	LPCSTR nextGroup = pszArgs;

	while (GetToken(nextGroup, group, &nextGroup, false))
	{
		for(LabelListIterator i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), group) == 0 || stricmp((*i)->labelGroup.c_str(), "*") == 0 )
				(*i)->hide();
		}
	}
}

void BangShortcutShow(HWND hwndCaller, LPCSTR pszArgs)
{
	char group[4096];
	LPCSTR nextGroup = pszArgs;

	while (GetToken(nextGroup, group, &nextGroup, false))
	{
		for(LabelListIterator i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), group) == 0 || stricmp((*i)->labelGroup.c_str(), "*") == 0 )
				(*i)->show();
		}
	}
}

void BangShortcutToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	char group[4096];
	LPCSTR nextGroup = pszArgs;

	while (GetToken(nextGroup, group, &nextGroup, false))
	{
		for(LabelListIterator i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), group) == 0 || stricmp((*i)->labelGroup.c_str(), "*") == 0 )
			{
				if((*i)->isVisible())
					(*i)->hide();
				else
					(*i)->show();
			}
		}
	}
}

void BangShortcutOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
	char group[4096];
	LPCSTR nextGroup = pszArgs;

	while (GetToken(nextGroup, group, &nextGroup, false))
	{
		for(LabelListIterator i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), group) == 0 || stricmp((*i)->labelGroup.c_str(), "*") == 0 )
				(*i)->setAlwaysOnTop(true);
		}
	}
}

void BangShortcutOnBottom(HWND hwndCaller, LPCSTR pszArgs)
{
	char group[4096];
	LPCSTR nextGroup = pszArgs;

	while (GetToken(nextGroup, group, &nextGroup, false))
	{
		for(LabelListIterator i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), group) == 0 || stricmp((*i)->labelGroup.c_str(), "*") == 0 )
				(*i)->setAlwaysOnTop(false);
		}
	}
}

void BangShortcutOnTopToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	char group[4096];
	LPCSTR nextGroup = pszArgs;

	while (GetToken(nextGroup, group, &nextGroup, false))
	{
		for(LabelListIterator i = shortcutlabelList.begin(); i != shortcutlabelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), group) == 0 || stricmp((*i)->labelGroup.c_str(), "*") == 0 )
				(*i)->setAlwaysOnTop(!(*i)->isAlwaysOnTop());
		}
	}
}

//Parse Current Evars In Bang
//!ParseEvars !anybang with evars escape code '%#'
void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	string result = pszArgs;
	while ( strstr(result.c_str(), "%#") != NULL )
	{
		result.replace(result.find("%#"), 2, "$", 1);
	}
	LSExecute(NULL, result.c_str(), SW_SHOWNORMAL);
}

//Set Evar
//!SetEvar "'$'evar'$'" "value"
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	char evar[64], value[128];
	char *buffers[] = {evar, value};

	if(LCTokenize(pszArgs, buffers, 2, NULL) >= 2)
	{
		PrefixedLSSetVariable( (string)evar, "", (string)value );
	}
}